/*
 * Copyright (C) 2024  Fluence DAO
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package aqua.definitions

import aqua.definitions.*
import aqua.res.FuncRes
import aqua.types.*

import io.circe.*
import io.circe.parser.*
import io.circe.syntax.*
import scala.annotation.tailrec

// Represents the Aqua types
sealed trait TypeDefinition {
  def tag: String
}

object TypeDefinition {

  given Encoder[ProductTypeDef] = {
    case d @ LabeledProductTypeDef(fields) =>
      Json.obj(
        ("tag", Json.fromString(d.tag)),
        ("fields", Json.fromFields(fields.map { case (n, t) => (n, t.asJson) }))
      )
    case d @ UnlabeledProductTypeDef(items) =>
      Json.obj(
        ("tag", Json.fromString(d.tag)),
        ("items", Json.fromValues(items.map(_.asJson)))
      )
    case d @ NilTypeDef =>
      Json.obj(
        ("tag", Json.fromString(d.tag))
      )
  }

  given Encoder[TypeDefinition] = {
    case d @ ScalarTypeDef(name) =>
      Json.obj(
        ("tag", Json.fromString(d.tag)),
        ("name", Json.fromString(name))
      )
    case d @ ArrayTypeDef(t) =>
      Json.obj(
        ("tag", Json.fromString(d.tag)),
        ("type", t.asJson)
      )
    case d @ StructTypeDef(name, fields) =>
      Json.obj(
        ("tag", Json.fromString(d.tag)),
        ("name", Json.fromString(name)),
        ("fields", Json.fromFields(fields.toList.map { case (n, t) => (n, t.asJson) }))
      )

    case d @ OptionTypeDef(t) =>
      Json.obj(
        ("tag", Json.fromString(d.tag)),
        ("type", t.asJson)
      )
    case d @ ArrowTypeDef(domain, codomain) =>
      Json.obj(
        ("tag", Json.fromString(d.tag)),
        ("domain", domain.asJson),
        ("codomain", codomain.asJson)
      )
    case d =>
      Json.obj(
        ("tag", Json.fromString(d.tag))
      )
  }

  given Encoder[ServiceDef] = { case ServiceDef(sId, functions, name) =>
    Json.obj(
      ("defaultServiceId", sId.asJson),
      ("functions", (functions: ProductTypeDef).asJson)
    )
  }

  given Encoder[NamesConfig] = { case n: NamesConfig =>
    import n.*
    Json.obj(
      ("relay", Json.fromString(relay)),
      ("getDataSrv", Json.fromString(getDataSrv)),
      ("callbackSrv", Json.fromString(callbackSrv)),
      ("responseSrv", Json.fromString(responseSrv)),
      ("responseFnName", Json.fromString(responseFnName)),
      ("errorHandlingSrv", Json.fromString(errorHandlingSrv)),
      ("errorFnName", Json.fromString(errorFnName))
    )
  }

  given Encoder[FunctionDef] = { case FunctionDef(fName, arrow, names) =>
    Json.obj(
      ("functionName", Json.fromString(fName)),
      ("arrow", (arrow: TypeDefinition).asJson),
      ("names", names.asJson)
    )
  }

  def apply(t: Option[Type]): TypeDefinition = t.map(apply).getOrElse(NilTypeDef)

  def apply(t: Type): TypeDefinition = {
    t match {
      case OptionType(t) =>
        OptionTypeDef(TypeDefinition(t))
      case t: CollectionType => ArrayTypeDef(TypeDefinition(t.element))
      case StructType(name, fields) =>
        StructTypeDef(name, fields.toSortedMap.view.mapValues(TypeDefinition.apply).toMap)
      case AbilityType(name, fieldAndArrows) =>
        // TODO: change in union with JS side
        StructTypeDef(name, fieldAndArrows.toSortedMap.view.mapValues(TypeDefinition.apply).toMap)
      case t: ScalarType => ScalarTypeDef.fromScalar(t)
      case t: LiteralType => ScalarTypeDef.fromLiteral(t)
      case t: ProductType => ProductTypeDef(t)
      case t: ArrowType =>
        ArrowTypeDef(t)
      case TopType => TopTypeDef
      case BottomType => BottomTypeDef
    }
  }
}

sealed trait ProductTypeDef extends TypeDefinition

object ProductTypeDef {

  def apply(t: ProductType): ProductTypeDef = {
    t match {
      case lt: LabeledConsType =>
        LabeledProductTypeDef(
          lt.toLabelledList().map { case (n, t) => (n, TypeDefinition(t)) }
        )
      case ut: UnlabeledConsType =>
        UnlabeledProductTypeDef(ut.toList.map(TypeDefinition.apply))
      case NilType => NilTypeDef
    }
  }
}

case class OptionTypeDef(`type`: TypeDefinition) extends TypeDefinition { val tag = "option" }
case class ScalarTypeDef private (name: String) extends TypeDefinition { val tag = "scalar" }

object ScalarTypeDef {
  def fromScalar(s: ScalarType): ScalarTypeDef = ScalarTypeDef(s.name)
  def fromLiteral(s: LiteralType): ScalarTypeDef = ScalarTypeDef(s.name)
}

case class ArrayTypeDef(`type`: TypeDefinition) extends TypeDefinition { val tag = "array" }

case class ArrowTypeDef(
  domain: ProductTypeDef,
  codomain: ProductTypeDef
) extends TypeDefinition {
  val tag = "arrow"
}

object ArrowTypeDef {

  def apply(at: ArrowType): ArrowTypeDef =
    ArrowTypeDef(ProductTypeDef(at.domain), ProductTypeDef(at.codomain))
}

case object NilTypeDef extends ProductTypeDef {
  val tag = "nil"
}

case object TopTypeDef extends TypeDefinition {
  val tag = "topType"
}

case object BottomTypeDef extends TypeDefinition {
  val tag = "bottomType"
}

case class StructTypeDef(name: String, fields: Map[String, TypeDefinition]) extends TypeDefinition {
  val tag = "struct"
}

case class LabeledProductTypeDef(fields: List[(String, TypeDefinition)]) extends ProductTypeDef {
  val tag = "labeledProduct"

  override def toString: String = s"LabeledProduct(${fields.map(_.toString())})"
}

case class UnlabeledProductTypeDef(items: List[TypeDefinition]) extends ProductTypeDef {
  val tag = "unlabeledProduct"
}

// Names of services and functions that are used in air
case class NamesConfig(
  relay: String,
  getDataSrv: String,
  callbackSrv: String,
  responseSrv: String,
  responseFnName: String,
  errorHandlingSrv: String,
  errorFnName: String
)

// Describes service
case class ServiceDef(
  defaultServiceId: Option[String],
  functions: LabeledProductTypeDef,
  name: String
)

// Describes top-level function
case class FunctionDef(
  functionName: String,
  arrow: ArrowTypeDef,
  names: NamesConfig
)

object FunctionDef {

  def apply(func: FuncRes): FunctionDef = {
    val args = LabeledProductTypeDef(func.args.map(a => (a.name, TypeDefinition(a.`type`))))

    val names = NamesConfig(
      func.relayVarName.getOrElse("-relay-"),
      func.dataServiceId,
      func.callbackServiceId,
      func.callbackServiceId,
      func.respFuncName,
      func.errorHandlerId,
      func.errorFuncName
    )
    FunctionDef(
      func.funcName,
      ArrowTypeDef(args, ProductTypeDef(func.returnType)),
      names
    )
  }
}
